VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "Resampler"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Private Type src_data
    sng_data_in_ptr     As Long
    sng_data_out_ptr    As Long
    input_frames        As Long
    output_frames       As Long
    input_frames_used   As Long
    output_frames_gen   As Long
    end_of_input        As Long
    filler              As Long
    src_ratio           As Double
End Type

Private Enum src_result
    SRC_ERR_NO_ERROR = 0
    SRC_ERR_MALLOC_FAILED
    SRC_ERR_BAD_STATE
    SRC_ERR_BAD_DATA
    SRC_ERR_BAD_DATA_PTR
    SRC_ERR_NO_PRIVATE
    SRC_ERR_BAD_SRC_RATIO
    SRC_ERR_BAD_PROC_PTR
    SRC_ERR_SHIFT_BITS
    SRC_ERR_FILTER_LEN
    SRC_ERR_BAD_CONVERTER
    SRC_ERR_BAD_CHANNEL_COUNT
    SRC_ERR_SINC_BAD_BUFFER_LEN
    SRC_ERR_SIZE_INCOMPATIBILITY
    SRC_ERR_BAD_PRIV_PTR
    SRC_ERR_BAD_SINC_STATE
    SRC_ERR_DATA_OVERLAP
    SRC_ERR_BAD_CALLBACK
    SRC_ERR_BAD_MODE
    SRC_ERR_NULL_CALLBACK
    SRC_ERR_MAX_ERROR
End Enum

Public Enum rsmp_method
    RSMP_SINC_BEST_QUALITY = 0
    RSMP_SINC_MEDIUM_QUALITY = 1
    RSMP_SINC_FASTEST = 2
    RSMP_ZERO_ORDER_HOLD = 3
    RSMP_LINEAR = 4
End Enum

Public Enum rsmp_result
    RSMP_ERR_NO_ERROR = 0
    RSMP_ERR_INVALID_CHANNELS
    RSMP_ERR_INVALID_RATIO
    RSMP_ERR_UNKNOWN
End Enum

Private clsSRC              As clsDLL

Private lngInpSamplerate    As Long
Private lngOutSamplerate    As Long
Private intInpChannels      As Integer
Private intOutChannels      As Integer
Private udtData             As src_data
Private hState              As Long

Public Property Get IsRunning( _
) As Boolean

    IsRunning = lngInpSamplerate <> 0
End Property

Public Property Get InputSamplerate( _
) As Long

    InputSamplerate = lngInpSamplerate
End Property

Public Property Get OutputSamplerate( _
) As Long

    OutputSamplerate = lngOutSamplerate
End Property

Public Property Get InputChannels( _
) As Integer

    InputChannels = intInpChannels
End Property

Public Property Get OutputChannels( _
) As Integer

    OutputChannels = intOutChannels
End Property

Public Sub CloseConverter()
    ZeroMem udtData, Len(udtData)
    hState = src_delete(hState)
    lngInpSamplerate = 0
    lngOutSamplerate = 0
    intInpChannels = 0
    intOutChannels = 0
End Sub

Public Function Resample( _
    ByVal short_buffer_ptr As Long, _
    ByVal Length As Long, _
    ByRef output() As Integer, _
    ByRef outlength As Long _
) As Boolean

    Dim sngInp()    As Single
    Dim sngOut()    As Single
    Dim lngOutLen   As Long
    Dim udeErr      As src_result

    lngOutLen = Length * udtData.src_ratio * 3

    ReDim sngInp(Length - 1) As Single
    ReDim sngOut(lngOutLen - 1) As Single

    src_short_to_float_array short_buffer_ptr, _
                             VarPtr(sngInp(0)), _
                             Length

    udtData.input_frames = Length / intInpChannels
    udtData.output_frames = lngOutLen / intInpChannels
    udtData.sng_data_in_ptr = VarPtr(sngInp(0))
    udtData.sng_data_out_ptr = VarPtr(sngOut(0))

    udeErr = src_process(hState, udtData)
    If udeErr <> SRC_ERR_NO_ERROR Then
        Debug.Print "Resample: src_process Fehler " & udeErr
        Exit Function
    End If

    lngOutLen = udtData.output_frames_gen * intInpChannels
    ReDim output(lngOutLen - 1) As Integer

    src_float_to_short_array VarPtr(sngOut(0)), _
                             VarPtr(output(0)), _
                             lngOutLen

    If intInpChannels > intOutChannels Then
        StereoToMono output, lngOutLen
        lngOutLen = lngOutLen / 2
    ElseIf intInpChannels < intOutChannels Then
        MonoToStereo output, lngOutLen
        lngOutLen = lngOutLen * 2
    End If

    outlength = lngOutLen

    Resample = True
End Function

Public Function InitConversion( _
    ByVal input_samplerate As Long, _
    ByVal output_samplerate As Long, _
    ByVal input_channels As Integer, _
    ByVal output_channels As Integer, _
    ByVal resample_method As rsmp_method _
) As rsmp_result

    Dim dblRatio    As Double
    Dim udeErr      As src_result

    CloseConverter

    dblRatio = output_samplerate / input_samplerate

    If Not src_is_valid_ratio(dblRatio) Then
        InitConversion = RSMP_ERR_INVALID_RATIO
        Exit Function
    End If

    If input_channels < 1 Or input_channels > 2 Then
        InitConversion = RSMP_ERR_INVALID_CHANNELS
        Exit Function
    End If

    If output_channels < 1 Or output_channels > 2 Then
        InitConversion = RSMP_ERR_INVALID_CHANNELS
        Exit Function
    End If

    lngInpSamplerate = input_samplerate
    lngOutSamplerate = output_samplerate
    intInpChannels = input_channels
    intOutChannels = output_channels

    udtData.src_ratio = dblRatio

    hState = src_new(resample_method, _
                     intInpChannels, _
                     udeErr)

    If hState = 0 Then
        Debug.Print "InitConversion: src_new Fehler " & udeErr
        InitConversion = RSMP_ERR_UNKNOWN
        Exit Function
    End If

    InitConversion = RSMP_ERR_NO_ERROR
End Function

Private Sub MonoToStereo( _
    buf() As Integer, _
    ByVal Length As Long _
)

    Dim i       As Long
    Dim out() As Integer

    ReDim out(Length * 2 - 1) As Integer

    For i = 0 To Length - 1
        out(i * 2 + 0) = buf(i)
        out(i * 2 + 1) = buf(i)
    Next

    buf = out
End Sub

Private Sub StereoToMono( _
    buf() As Integer, _
    ByVal Length As Long _
)

    Dim i   As Long

    For i = 0 To Length / 2 - 1
        buf(i) = cut_sample(CLng((buf(i * 2)) + CLng(buf(i * 2 + 1))) / 2&)
    Next

    ReDim Preserve buf(Length / 2 - 1) As Integer
End Sub

Private Function cut_sample( _
    ByVal lngSample As Long _
) As Integer

    If lngSample > 32767 Then
        cut_sample = 32767
    ElseIf lngSample < -32768 Then
        cut_sample = -32768
    Else
        cut_sample = lngSample
    End If
End Function

'''''''''''''''''''''''''''''''''''''''''''''''
' libsamplerate Wrapper
'''''''''''''''''''''''''''''''''''''''''''''''

Private Function src_init_lib( _
) As Boolean

    If clsSRC Is Nothing Then
        Set clsSRC = New clsDLL
    End If

    If clsSRC.ModuleHandle = 0 Then
        src_init_lib = clsSRC.LoadDLL("libsamplerate.dll")
        clsSRC.IsCDECL = True
    End If
End Function

Private Sub src_unload_lib()
    If Not clsSRC Is Nothing Then
        clsSRC.UnloadDLL
    End If
End Sub

Private Function src_new( _
    ByVal converter_type As rsmp_method, _
    ByVal Channels As Long, _
    ByRef src_err As src_result _
) As Long

    src_new = clsSRC.CallFunc("src_new", _
                              converter_type, _
                              Channels, _
                              VarPtr(src_err))
End Function

Private Function src_delete( _
    ByVal src_state As Long _
) As Long

    src_delete = clsSRC.CallFunc("src_delete", _
                                 src_state)
End Function

Private Function src_process( _
    ByVal src_state As Long, _
    src_data As src_data _
) As Long

    src_process = clsSRC.CallFunc("src_process", _
                                  src_state, _
                                  VarPtr(src_data))
End Function

Private Function src_set_ratio( _
    ByVal src_state As Long, _
    ByVal new_ratio As Double _
) As Long

    Dim loDouble    As Long
    Dim hiDouble    As Long

    CpyMem hiDouble, ByVal VarPtr(new_ratio) + 0, 4
    CpyMem loDouble, ByVal VarPtr(new_ratio) + 4, 4

    src_set_ratio = clsSRC.CallFunc("src_set_ratio", _
                                    src_state, _
                                    hiDouble, _
                                    loDouble)
End Function

Private Function src_reset( _
    ByVal src_state As Long _
) As Long

    src_reset = clsSRC.CallFunc("src_reset", _
                                src_state)
End Function

Private Function src_is_valid_ratio( _
    ByVal ratio As Double _
) As Boolean

    Dim loDouble    As Long
    Dim hiDouble    As Long

    CpyMem hiDouble, ByVal VarPtr(ratio) + 0, 4
    CpyMem loDouble, ByVal VarPtr(ratio) + 4, 4

    src_is_valid_ratio = clsSRC.CallFunc("src_is_valid_ratio", _
                                         hiDouble, _
                                         loDouble)
End Function

Private Function src_error( _
    ByVal src_state As Long _
) As src_result

    src_error = clsSRC.CallFunc("src_error", _
                                src_state)
End Function

Private Function src_simple( _
    src_data As src_data, _
    ByVal converter_type As rsmp_method, _
    ByVal Channels As Long _
) As Long

    src_simple = clsSRC.CallFunc("src_simple", _
                                 VarPtr(src_data), _
                                 converter_type, _
                                 Channels)
End Function

Private Sub src_short_to_float_array( _
    ByVal short_arr_in_ptr As Long, _
    ByVal float_arr_out_ptr As Long, _
    ByVal Length As Long _
)

    clsSRC.CallFunc "src_short_to_float_array", _
                    short_arr_in_ptr, _
                    float_arr_out_ptr, _
                    Length
End Sub

Private Sub src_float_to_short_array( _
    ByVal float_arr_in_ptr As Long, _
    ByVal short_arr_out_ptr As Long, _
    ByVal Length As Long _
)

    clsSRC.CallFunc "src_float_to_short_array", _
                    float_arr_in_ptr, _
                    short_arr_out_ptr, _
                    Length
End Sub

Private Sub Class_Initialize()
    src_init_lib
End Sub

Private Sub Class_Terminate()
    src_unload_lib
End Sub
