VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "StreamAPE"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Implements ISoundStream
Implements ISoundInfo


Private Const Extensions    As String = "ape"
Private Const Description   As String = "Monkey's Audio"


Private Declare Function APEDecompress_Create Lib "MACDll.dll" _
Alias "c_APEDecompress_Create" ( _
    ByVal pFilename As String, _
    ByVal pErrorCode As Long _
) As APE_HANDLE

Private Declare Sub APEDecompress_Destroy Lib "MACDll.dll" _
Alias "c_APEDecompress_Destroy" ( _
    ByVal handle As APE_HANDLE _
)

Private Declare Function APEDecompress_GetInfo Lib "MACDll.dll" _
Alias "c_APEDecompress_GetInfo" ( _
    ByVal handle As APE_HANDLE, _
    ByVal info As APE_DECOMPRESS_FIELDS, _
    ByVal nParam1 As Long, _
    ByVal nParam2 As Long _
) As Long

Private Declare Function APEDecompress_Seek Lib "MACDll.dll" _
Alias "c_APEDecompress_Seek" ( _
    ByVal handle As APE_HANDLE, _
    ByVal nBlockOffset As Long _
) As Long

Private Declare Function APEDecompress_GetData Lib "MACDll.dll" _
Alias "c_APEDecompress_GetData" ( _
    ByVal handle As APE_HANDLE, _
    ByVal pBuffer As Long, _
    ByVal intBlocks As Long, _
    ByRef pBlocksRetrieved As Long _
) As Long

Private Type APE_TAG_FOOTER
    preample        As String * 8
    tagver          As Long
    tagsize         As Long
    cTags           As Long
    flags           As Long
    zero1           As Long
    zero2           As Long
End Type

Private Type APE_TAG_ITEM
    size            As Long
    flags           As Long
End Type

Private Enum APE_HANDLE
    ape_invalid_handle = 0
End Enum

Private Enum APE_COMPRESSION_LEVELS
    COMPRESSION_LEVEL_FAST = 1000
    COMPRESSION_LEVEL_NORMAL = 2000
    COMPRESSION_LEVEL_HIGH = 3000
    COMPRESSION_LEVEL_EXTRA_HIGH = 4000
    COMPRESSION_LEVEL_INSANE = 5000
End Enum

Private Enum APE_DECOMPRESS_FIELDS
    APE_INFO_FILE_VERSION = 1000            ' version of the APE file * 1000 (3.93 = 3930) [ignored, ignored]
    APE_INFO_COMPRESSION_LEVEL = 1001       ' compression level of the APE file [ignored, ignored]
    APE_INFO_FORMAT_FLAGS = 1002            ' format flags of the APE file [ignored, ignored]
    APE_INFO_SAMPLE_RATE = 1003             ' sample rate (Hz) [ignored, ignored]
    APE_INFO_BITS_PER_SAMPLE = 1004         ' bits per sample [ignored, ignored]
    APE_INFO_BYTES_PER_SAMPLE = 1005        ' number of bytes per sample [ignored, ignored]
    APE_INFO_CHANNELS = 1006                ' channels [ignored, ignored]
    APE_INFO_BLOCK_ALIGN = 1007             ' block alignment [ignored, ignored]
    APE_INFO_BLOCKS_PER_FRAME = 1008        ' number of blocks in a frame (frames are used internally)  [ignored, ignored]
    APE_INFO_FINAL_FRAME_BLOCKS = 1009      ' blocks in the final frame (frames are used internally) [ignored, ignored]
    APE_INFO_TOTAL_FRAMES = 1010            ' total number frames (frames are used internally) [ignored, ignored]
    APE_INFO_WAV_HEADER_BYTES = 1011        ' header bytes of the decompressed WAV [ignored, ignored]
    APE_INFO_WAV_TERMINATING_BYTES = 1012   ' terminating bytes of the decompressed WAV [ignored, ignored]
    APE_INFO_WAV_DATA_BYTES = 1013          ' data bytes of the decompressed WAV [ignored, ignored]
    APE_INFO_WAV_TOTAL_BYTES = 1014         ' total bytes of the decompressed WAV [ignored, ignored]
    APE_INFO_APE_TOTAL_BYTES = 1015         ' total bytes of the APE file [ignored, ignored]
    APE_INFO_TOTAL_BLOCKS = 1016            ' total blocks of audio data [ignored, ignored]
    APE_INFO_LENGTH_MS = 1017               ' length in ms (1 sec = 1000 ms) [ignored, ignored]
    APE_INFO_AVERAGE_BITRATE = 1018         ' average bitrate of the APE [ignored, ignored]
    APE_INFO_FRAME_BITRATE = 1019           ' bitrate of specified APE frame [frame index, ignored]
    APE_INFO_DECOMPRESSED_BITRATE = 1020    ' bitrate of the decompressed WAV [ignored, ignored]
    APE_INFO_PEAK_LEVEL = 1021              ' peak audio level (obsolete) (-1 is unknown) [ignored, ignored]
    APE_INFO_SEEK_BIT = 1022                ' bit offset [frame index, ignored]
    APE_INFO_SEEK_BYTE = 1023               ' byte offset [frame index, ignored]
    APE_INFO_WAV_HEADER_DATA = 1024         ' error code [buffer *, max bytes]
    APE_INFO_WAV_TERMINATING_DATA = 1025    ' error code [buffer *, max bytes]
    APE_INFO_WAVEFORMATEX = 1026            ' error code [waveformatex *, ignored]
    APE_INFO_IO_SOURCE = 1027               ' I/O source (CIO *) [ignored, ignored]
    APE_INFO_FRAME_BYTES = 1028             ' bytes (compressed) of the frame [frame index, ignored]
    APE_INFO_FRAME_BLOCKS = 1029            ' blocks in a given frame [frame index, ignored]
    APE_INFO_TAG = 1030                     ' point to tag (CAPETag *) [ignored, ignored]

    APE_DECOMPRESS_CURRENT_BLOCK = 2000     ' current block location [ignored, ignored]
    APE_DECOMPRESS_CURRENT_MS = 2001        ' current millisecond location [ignored, ignored]
    APE_DECOMPRESS_TOTAL_BLOCKS = 2002      ' total blocks in the decompressors range [ignored, ignored]
    APE_DECOMPRESS_LENGTH_MS = 2003         ' total blocks in the decompressors range [ignored, ignored]
    APE_DECOMPRESS_CURRENT_BITRATE = 2004   ' current bitrate [ignored, ignored]
    APE_DECOMPRESS_AVERAGE_BITRATE = 2005   ' average bitrate (works with ranges) [ignored, ignored]

    APE_INTERNAL_INFO = 3000                ' for internal use -- don't use (returns APE_FILE_INFO *) [ignored, ignored]
End Enum

Private hAPE                    As APE_HANDLE

Private lngSamplerate           As Long
Private intChannels             As Integer
Private lngBitrate              As Long
Private intBitsPerSample        As Integer
Private lngLength               As Long
Private lngBlocks               As Long
Private lngBlockAlign           As Long

Private Const BUFFER_SIZE_MS    As Long = 50

Private intBuffer()             As Integer

Private lngPosInBuffer          As Long
Private lngBufferData           As Long

Private blnEOS                  As Boolean

Private clsTags                 As StreamTags

''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''

Private Sub Class_Initialize()
    Set clsTags = New StreamTags
End Sub

Private Sub Class_Terminate()
    ISoundStream_StreamClose
End Sub

''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''

Private Property Get ISoundInfo_BitsPerSample( _
) As Integer

    ISoundInfo_BitsPerSample = 16
End Property

Private Property Get ISoundInfo_BitsPerSecond( _
) As Long

    ISoundInfo_BitsPerSecond = lngBitrate * 1000
End Property

Private Property Get ISoundInfo_channels( _
) As Integer

    ISoundInfo_channels = intChannels
End Property

Private Property Get ISoundInfo_Duration( _
) As Long

    ISoundInfo_Duration = lngLength
End Property

Private Property Get ISoundInfo_Position( _
) As Long

    Dim lngRet As Long

    lngRet = APEDecompress_GetInfo(hAPE, _
                                   APE_DECOMPRESS_CURRENT_MS, _
                                   0, 0)

    ISoundInfo_Position = lngRet
End Property

Private Property Get ISoundInfo_SamplesPerSecond( _
) As Long

    ISoundInfo_SamplesPerSecond = lngSamplerate
End Property

Private Property Get ISoundInfo_Tags( _
) As StreamTags

    Set ISoundInfo_Tags = clsTags
End Property

Private Property Get ISoundStream_Description( _
) As String

    ISoundStream_Description = Description
End Property

''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''

Private Property Get ISoundStream_StreamInfo( _
) As ISoundInfo

    Set ISoundStream_StreamInfo = Me
End Property

Private Property Get ISoundStream_EndOfStream( _
) As Boolean

    ISoundStream_EndOfStream = blnEOS
End Property

Private Function ISoundStream_Extensions( _
) As String()

    ISoundStream_Extensions = Split(Extensions, ";")
End Function

Private Function ISoundStream_StreamClose( _
) As SND_RESULT

    If hAPE = ape_invalid_handle Then
        ISoundStream_StreamClose = SND_ERR_INVALID_SOURCE
    Else
        APEDecompress_Destroy hAPE
        hAPE = ape_invalid_handle

        blnEOS = False

        clsTags.ClearTags

        ISoundStream_StreamClose = SND_ERR_SUCCESS
    End If
End Function

Private Function ISoundStream_QuickStreamInfo( _
    ByVal Source As String, _
    info As ISoundInfo _
) As SND_RESULT


    Dim lngErr      As Long
    Dim hAPEQ       As APE_HANDLE
    Dim clsInfo     As QuickStreamInfo
    Dim clsTagsQ    As StreamTags

    Set clsInfo = New QuickStreamInfo
    Set clsTagsQ = New StreamTags
    Set info = clsInfo

    If Not IsValidFile("MACDll.dll") Then
        If Not IsValidFile(GetSystemDirectory & "MACDll.dll") Then
            ISoundStream_QuickStreamInfo = SND_ERR_INTERNAL
            Exit Function
        End If
    End If

    If Not IsValidFile(Source) Then
        ISoundStream_QuickStreamInfo = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    hAPEQ = APEDecompress_Create(Source, lngErr)
    If lngErr <> 0 Then
        ISoundStream_QuickStreamInfo = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    If APEDecompress_GetInfo(hAPEQ, APE_INFO_BITS_PER_SAMPLE, 0, 0) = 24 Then
        APEDecompress_Destroy hAPEQ
        ISoundStream_QuickStreamInfo = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    ReadAPETags Source, clsTagsQ

    ' Datei Infos
    clsInfo.SetValues APEDecompress_GetInfo(hAPEQ, APE_INFO_SAMPLE_RATE, 0, 0), _
                      APEDecompress_GetInfo(hAPEQ, APE_INFO_CHANNELS, 0, 0), _
                      16, _
                      APEDecompress_GetInfo(hAPEQ, APE_INFO_AVERAGE_BITRATE, 0, 0), _
                      APEDecompress_GetInfo(hAPEQ, APE_INFO_LENGTH_MS, 0, 0), _
                      0, _
                      clsTagsQ

    APEDecompress_Destroy hAPEQ

    ISoundStream_QuickStreamInfo = SND_ERR_SUCCESS
End Function

Private Function ISoundStream_StreamOpen( _
    ByVal Source As String _
) As SND_RESULT

    Dim lngErr  As Long

    If Not IsValidFile("MACDll.dll") Then
        If Not IsValidFile(GetSystemDirectory & "MACDll.dll") Then
            ISoundStream_StreamOpen = SND_ERR_INTERNAL
            Exit Function
        End If
    End If

    ISoundStream_StreamClose

    If Not IsValidFile(Source) Then
        ISoundStream_StreamOpen = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    hAPE = APEDecompress_Create(Source, lngErr)
    If lngErr <> 0 Then
        ISoundStream_StreamOpen = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    intBitsPerSample = APEDecompress_GetInfo(hAPE, APE_INFO_BITS_PER_SAMPLE, 0, 0)
    lngBlockAlign = APEDecompress_GetInfo(hAPE, APE_INFO_BLOCK_ALIGN, 0, 0)
    lngSamplerate = APEDecompress_GetInfo(hAPE, APE_INFO_SAMPLE_RATE, 0, 0)
    intChannels = APEDecompress_GetInfo(hAPE, APE_INFO_CHANNELS, 0, 0)
    lngBitrate = APEDecompress_GetInfo(hAPE, APE_INFO_AVERAGE_BITRATE, 0, 0)
    lngLength = APEDecompress_GetInfo(hAPE, APE_INFO_LENGTH_MS, 0, 0)
    lngBlocks = APEDecompress_GetInfo(hAPE, APE_INFO_TOTAL_BLOCKS, 0, 0)

    If intBitsPerSample <> 8 Then
        If intBitsPerSample <> 16 Then
            ISoundStream_StreamClose
            ISoundStream_StreamOpen = SND_ERR_INVALID_SOURCE
            Exit Function
        End If
    End If

    ReDim intBuffer(BUFFER_SIZE_MS * lngSamplerate / 1000 - 1) As Integer

    ReadAPETags Source, clsTags

    ISoundStream_StreamOpen = SND_ERR_SUCCESS
End Function

Private Function ISoundStream_StreamRead( _
    ByVal buffer_ptr As Long, _
    ByVal buffer_len As Long, _
    buffer_read As Long _
) As SND_RESULT

    ISoundStream_StreamRead = SND_ERR_SUCCESS

    buffer_read = 0

    Do While buffer_read < buffer_len

        If lngBufferData = 0 Then
            If Not FillBuffer Then
                ISoundStream_StreamRead = SND_ERR_END_OF_STREAM
                Exit Function
            End If

        ElseIf (lngBufferData - lngPosInBuffer) < (buffer_len - buffer_read) Then
            If 0 < (lngBufferData - lngPosInBuffer) Then

                If 0 = IsBadReadPtr(ByVal VarPtr(intBuffer(0)) + lngPosInBuffer, _
                                    lngBufferData - lngPosInBuffer) Then

                    If 0 = IsBadWritePtr(ByVal buffer_ptr + buffer_read, _
                                         lngBufferData - lngPosInBuffer) Then

                        CpyMem ByVal buffer_ptr + buffer_read, _
                               ByVal VarPtr(intBuffer(0)) + lngPosInBuffer, _
                               lngBufferData - lngPosInBuffer

                    End If
                End If

                buffer_read = buffer_read + (lngBufferData - lngPosInBuffer)
            End If

            If Not FillBuffer Then
                ISoundStream_StreamRead = SND_ERR_END_OF_STREAM
                Exit Function
            End If

        Else
            If 0 = IsBadReadPtr(ByVal VarPtr(intBuffer(0)) + lngPosInBuffer, _
                                buffer_len - buffer_read) Then

                If 0 = IsBadWritePtr(ByVal buffer_ptr + buffer_read, _
                                     buffer_len - buffer_read) Then

                    CpyMem ByVal buffer_ptr + buffer_read, _
                           ByVal VarPtr(intBuffer(0)) + lngPosInBuffer, _
                           buffer_len - buffer_read

                End If
            End If

            lngPosInBuffer = lngPosInBuffer + (buffer_len - buffer_read)
            buffer_read = buffer_read + (buffer_len - buffer_read)

        End If

    Loop
End Function

Private Function ISoundStream_StreamSeek( _
    ByVal Value As Long, _
    ByVal seek_mode As SND_SEEK_MODE _
) As SND_RESULT

    Dim lngRet      As Long
    Dim lngBlockPos As Long

    Select Case seek_mode

        Case SND_SEEK_PERCENT
            If Value < 0 Or Value > 99 Then
                ISoundStream_StreamSeek = SND_ERR_OUT_OF_RANGE
                Exit Function
            End If

            lngBlockPos = lngBlocks / 100 * Value

        Case SND_SEEK_SECONDS
            If Value < 0 Or Value > (ISoundInfo_Duration / 1000) Then
                ISoundStream_StreamSeek = SND_ERR_OUT_OF_RANGE
                Exit Function
            End If

            lngBlockPos = lngBlocks / 100 * (Value / (ISoundInfo_Duration / 1000) * 100)

    End Select

    lngRet = APEDecompress_Seek(hAPE, lngBlockPos)

    blnEOS = False

    ISoundStream_StreamSeek = SND_ERR_SUCCESS
End Function

''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''

Private Function FillBuffer( _
) As Boolean

    Dim lngRet          As Long
    Dim lngRetrieved    As Long

    Dim lngBytes        As Long
    Dim btBuffer()      As Byte

    lngBytes = (UBound(intBuffer) + 1&) * 2&

    ReDim btBuffer(lngBytes - 1) As Byte

    lngPosInBuffer = 0

    lngRet = APEDecompress_GetData(hAPE, _
                                   VarPtr(btBuffer(0)), _
                                   lngBytes / lngBlockAlign, _
                                   lngRetrieved)

    If lngRetrieved = 0 Then
        lngBufferData = 0
        blnEOS = True
        Exit Function
    End If

    lngBufferData = lngRetrieved * lngBlockAlign

    If intBitsPerSample = 8 Then

        intBuffer = bps_8_to_16(btBuffer, _
                                lngBufferData)

        lngBufferData = lngBufferData * 2

    Else

        CpyMem intBuffer(0), _
               btBuffer(0), _
               lngBufferData

    End If

    FillBuffer = True
End Function

Private Function bps_8_to_16( _
    btBuff8() As Byte, _
    ByVal Length As Long _
) As Integer()

    Dim i           As Long
    Dim intBuff16() As Integer

    ReDim intBuff16(Length - 1) As Integer

    For i = 0 To Length - 1
        intBuff16(i) = (btBuff8(i) - 128) * 256
    Next

    bps_8_to_16 = intBuff16
End Function

Private Sub ReadAPETags( _
    strFile As String, _
    tags As StreamTags _
)

    Dim fh          As Integer: fh = FreeFile

    Dim udtHdr      As APE_TAG_FOOTER
    Dim udtItem     As APE_TAG_ITEM

    Dim btTags()    As Byte
    Dim intPos      As Integer
    Dim j           As Integer
    Dim i           As Integer

    Dim strItemName As String
    Dim strItemVal  As String
    Dim Bt          As Byte

    tags.ClearTags

    Open strFile For Binary As #fh

        Seek #fh, LOF(fh) - Len(udtHdr) + 1
        Get #fh, , udtHdr

        If udtHdr.preample = "APETAGEX" Then

            ReDim btTags(udtHdr.tagsize - 1) As Byte
            Seek #fh, LOF(fh) - udtHdr.tagsize + 1
            Get #fh, , btTags

        End If

    Close #fh

    If Not udtHdr.preample = "APETAGEX" Then
        Exit Sub
    End If

    For i = 1 To udtHdr.cTags
        strItemName = ""
        strItemVal = ""

        CpyMem udtItem, btTags(intPos), Len(udtItem)
        intPos = intPos + 8

        Bt = 1
        Do While Not Bt = 0
            Bt = btTags(intPos)
            intPos = intPos + 1

            strItemName = strItemName & Chr$(Bt)
        Loop

        For j = intPos To intPos + udtItem.size - 1
            Bt = btTags(intPos)
            intPos = intPos + 1
            strItemVal = strItemVal & Chr$(Bt)
        Next

        strItemName = LCase$(RemNullChars(strItemName))

        If InStr(strItemVal, Chr$(0)) > 0 Then
            strItemVal = Left$(strItemVal, InStr(1, strItemVal, Chr$(0)) - 1)
        Else
            strItemVal = RemNullChars(strItemVal)
        End If

        Select Case strItemName
            Case "title", "artist", "album", "comment", "genre"
                tags.AddTag strItemName, strItemVal
            Case "track"
                If Not IsNumeric(strItemVal) Then
                    tags.AddTag "tracknumber", Str2Int(strItemVal)
                Else
                    tags.AddTag "tracknumber", strItemVal
                End If
             Case "year"
                If IsNumeric(strItemVal) Then
                    tags.AddTag "year", strItemVal
                Else
                    'tags.AddTag "year", ISO8601_to_Str(strItemVal)
                End If
        End Select
    Next
End Sub

Private Function Str2Int( _
    ByVal strVal As String _
) As Integer

    Dim intVal  As Integer
    Dim btHi    As Byte
    Dim btLo    As Byte

    btHi = Asc(Mid$(strVal, 1, 1))
    btLo = Asc(Mid$(strVal, 2, 1))

    CpyMem ByVal VarPtr(intVal) + 0, btHi, 1
    CpyMem ByVal VarPtr(intVal) + 1, btLo, 1

    Str2Int = intVal
End Function
