VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "StreamCDA"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Implements ISoundStream
Implements ISoundInfo


Private Const Description   As String = "CD Audio"
Private Const Extensions    As String = "cda"


Private Const CD_SAMPLERATE         As Long = 44100
Private Const CD_CHANNELS           As Long = 2
Private Const CD_BPS                As Long = 16
Private Const CD_BLOCKALIGN         As Long = 4

Private Const WAV_RIFF              As Long = 1179011410
Private Const WAV_CDDA              As Long = 1094992963
Private Const WAV_DATA              As Long = 1635017060
Private Const WAV_FMT               As Long = 544501094

Private Type CDAHdr
    RIFF                            As Long    '/* RIFF Chunk
    hdrlen                          As Long    '/* Header Length
    CDDA                            As Long    '/* CDDA Format
    fmt                             As Long    '/* fmt Chunk
    something                       As Long    '/* ?
End Type

Private Type CDAFileHeader
    hdr                             As CDAHdr '/* CDA Header
    ver                             As Integer '/* Version
    NumTrk                          As Integer '/* Track Number
    CD_ID                           As Long    '/* Disc ID
    TrkStartHSG                     As Long    '/* Track Start  (HSG)
    TrkLenHSG                       As Long    '/* Track Length (HSG)
    TrkStartRBook(3)                As Byte    '/* Track Start  (MSF)
    TrkLenRBook(3)                  As Byte    '/* Track Length (MSF)
End Type

Private Const SPR                   As Long = 20

Private clsToc                      As CDTOC
Private clsCurTrack                 As CDTOCTrack

Private lngBufferLen                As Long
Private lngBufferPos                As Long

Private lngLBA                      As Long
Private lngLBAEnd                   As Long

Private btBuffer(SPR * 32767 - 1&)  As Byte

Public Function GetDiscID( _
) As String

    Dim strID       As String
    Dim numtracks   As Integer

    If clsToc.TrackCount = 0 Then Exit Function

    With clsToc
        numtracks = .track(clsToc.TrackCount - 1).TrackNumber - .track(1).TrackNumber + 1
    End With

    strID = Hex$(cddb_discid(numtracks))
    strID = Left(String(8, "0"), 8 - Len(strID)) & strID

    GetDiscID = strID
End Function

Private Function cddb_sum( _
    ByVal n As Long _
) As Long

    Dim ret As Long

    Do While (n > 0)
        ret = UnsignedAdd(ret, n Mod 10)
        n = n \ 10
    Loop

    cddb_sum = ret
End Function

Private Function cddb_discid( _
    ByVal tot_tracks As Long _
) As Long

    Dim i       As Long
    Dim n       As Long
    Dim t       As Long
    Dim msfTc   As MSF
    Dim msfT0   As MSF
    Dim msfTl   As MSF

    msfT0 = LBA2MSF(clsToc.track(1).StartLBA)
    msfTl = LBA2MSF(clsToc.track(tot_tracks + 1).StartLBA)

    For i = 1 To tot_tracks
        msfTc = LBA2MSF(clsToc.track(i).StartLBA)
        n = UnsignedAdd(n, cddb_sum((msfTc.m * 60) + msfTc.S))
    Next

    t = ((msfTl.m * 60) + msfTl.S) - ((msfT0.m * 60) + msfT0.S)

    cddb_discid = (SHL(n Mod &HFF, 24) Or SHL(t, 8) Or tot_tracks)
End Function

Private Function UnsignedAdd( _
    start As Long, _
    Incr As Long _
) As Long

   If start And &H80000000 Then
      UnsignedAdd = start + Incr
   ElseIf (start Or &H80000000) < -Incr Then
      UnsignedAdd = start + Incr
   Else
      UnsignedAdd = (start + &H80000000) + (Incr + &H80000000)
   End If
End Function

Private Function FillBuffer( _
) As Boolean

    If Not ISoundStream_EndOfStream Then
        If Not clsCDAccess.ReadSectorsRAW(clsCurTrack.StartLBA + lngLBA, _
                                          SPR, VarPtr(btBuffer(0))) Then
            Exit Function
        End If

        lngLBA = lngLBA + SPR
        lngBufferPos = 0
        lngBufferLen = 2352& * SPR

        FillBuffer = True
    End If
End Function

Public Function ReadSectors( _
    ByVal LBA As Long, _
    ByVal sectors As Long, _
    ByVal buf_ptr As Long _
) As Boolean

    ReadSectors = clsCDAccess.ReadSectorsRAW(LBA, sectors, buf_ptr)
End Function

Public Property Get TrackCount( _
) As Long

    TrackCount = clsToc.TrackCount
End Property

Public Property Get TrackInfo( _
    ByVal index As Long _
) As CDTOCTrack

    Set TrackInfo = clsToc.track(index)
End Property

Public Function SelectDevice( _
    ByVal index As Long _
) As Boolean

    SelectDevice = clsCDAccess.SelectDevice(index)

    Set clsToc = New CDTOC
    clsCDAccess.GetTOC clsToc
End Function

Public Property Get DeviceCount( _
) As Long

    DeviceCount = clsCDAccess.DeviceCount
End Property

Public Property Get DeviceName( _
    ByVal index As Long _
) As String

    DeviceName = clsCDAccess.DeviceName(index)
End Property

Public Property Get DeviceChar( _
    ByVal index As Long _
) As String

    DeviceChar = clsCDAccess.DeviceChar(index)
End Property

Private Property Get ISoundInfo_BitsPerSample( _
) As Integer

    ISoundInfo_BitsPerSample = CD_BPS
End Property

Private Property Get ISoundInfo_BitsPerSecond( _
) As Long

    ISoundInfo_BitsPerSecond = CD_SAMPLERATE * CD_BLOCKALIGN * 8&
End Property

Private Property Get ISoundInfo_channels( _
) As Integer

    ISoundInfo_channels = CD_CHANNELS
End Property

Private Property Get ISoundInfo_Duration( _
) As Long

    ISoundInfo_Duration = (lngLBAEnd - clsCurTrack.StartLBA) / 75# * 1000#
End Property

Private Property Get ISoundInfo_Position( _
) As Long

    ISoundInfo_Position = lngLBA / 75# * 1000#
End Property

Private Property Get ISoundInfo_SamplesPerSecond( _
) As Long

    ISoundInfo_SamplesPerSecond = CD_SAMPLERATE
End Property

Private Property Get ISoundInfo_Tags( _
) As StreamTags

    Set ISoundInfo_Tags = New StreamTags
End Property

Private Property Get ISoundStream_Description( _
) As String

    ISoundStream_Description = Description
End Property

Private Property Get ISoundStream_EndOfStream( _
) As Boolean

    ISoundStream_EndOfStream = clsCurTrack.StartLBA + lngLBA >= lngLBAEnd
End Property

Private Function ISoundStream_Extensions( _
) As String()

    ISoundStream_Extensions = Split(Extensions, ";")
End Function

Private Function ISoundStream_QuickStreamInfo( _
    ByVal source As String, _
    info As ISoundInfo _
) As SND_RESULT

    Dim udtCDAHdr           As CDAFileHeader
    Dim hInp                As hFile
    Dim LBA                 As Long
    Dim LBALen              As Long
    Dim udtMSF              As MSF
    Dim clsQInfo            As QuickStreamInfo

    Set clsQInfo = New QuickStreamInfo
    Set info = clsQInfo

    If GetDirLevel(source) > 1 Then
        ISoundStream_QuickStreamInfo = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    hInp = FileOpen(source, GENERIC_READ, FILE_SHARE_READ)
    If hInp.handle = INVALID_HANDLE Then
        ISoundStream_QuickStreamInfo = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    FileRead hInp, VarPtr(udtCDAHdr), Len(udtCDAHdr)
    FileClose hInp

    If Not udtCDAHdr.hdr.RIFF = WAV_RIFF Then
        ISoundStream_QuickStreamInfo = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    If Not udtCDAHdr.hdr.fmt = WAV_FMT Then
        ISoundStream_QuickStreamInfo = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    If Not udtCDAHdr.hdr.CDDA = WAV_CDDA Then
        ISoundStream_QuickStreamInfo = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    With udtMSF
        .m = udtCDAHdr.TrkStartRBook(2)
        .S = udtCDAHdr.TrkStartRBook(1)
        .F = udtCDAHdr.TrkStartRBook(0)
    End With
    LBA = MSF2LBA(udtMSF)
    
    With udtMSF
        .m = udtCDAHdr.TrkLenRBook(2)
        .S = udtCDAHdr.TrkLenRBook(1)
        .F = udtCDAHdr.TrkLenRBook(0)
    End With
    LBALen = MSF2LBA(udtMSF)

    clsQInfo.SetValues 44100, 2, 16, _
                       44100 * 4& * 8&, _
                       (LBALen + 150) / 75# * 1000#, _
                       0, New StreamTags

    ISoundStream_QuickStreamInfo = SND_ERR_SUCCESS
End Function

Private Function ISoundStream_StreamClose( _
) As SND_RESULT

    If clsCurTrack Is Nothing Then
        ISoundStream_StreamClose = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    Set clsCurTrack = Nothing
    lngLBA = 0
    lngLBAEnd = 0
    lngBufferLen = 0
    lngBufferPos = 0

    ISoundStream_StreamClose = SND_ERR_SUCCESS
End Function

Private Property Get ISoundStream_StreamInfo( _
) As ISoundInfo

    Set ISoundStream_StreamInfo = Me
End Property

Private Function ISoundStream_StreamOpen( _
    ByVal source As String _
) As SND_RESULT

    Dim udtCDAHdr           As CDAFileHeader
    Dim hInp                As hFile
    Dim LBA                 As Long
    Dim LBALen              As Long
    Dim i                   As Long
    Dim j                   As Long
    Dim intFoundTrack       As Integer
    Dim udtMSF              As MSF
    Dim clsCurTOC           As CDTOC

    If GetDirLevel(source) > 1 Then
        ISoundStream_StreamOpen = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    hInp = FileOpen(source, GENERIC_READ, FILE_SHARE_READ)
    If hInp.handle = INVALID_HANDLE Then
        ISoundStream_StreamOpen = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    FileRead hInp, VarPtr(udtCDAHdr), Len(udtCDAHdr)
    FileClose hInp

    If Not udtCDAHdr.hdr.RIFF = WAV_RIFF Then
        ISoundStream_StreamOpen = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    If Not udtCDAHdr.hdr.fmt = WAV_FMT Then
        ISoundStream_StreamOpen = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    If Not udtCDAHdr.hdr.CDDA = WAV_CDDA Then
        ISoundStream_StreamOpen = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    With udtMSF
        .m = udtCDAHdr.TrkStartRBook(2)
        .S = udtCDAHdr.TrkStartRBook(1)
        .F = udtCDAHdr.TrkStartRBook(0)
    End With
    LBA = MSF2LBA(udtMSF)
    
    With udtMSF
        .m = udtCDAHdr.TrkLenRBook(2)
        .S = udtCDAHdr.TrkLenRBook(1)
        .F = udtCDAHdr.TrkLenRBook(0)
    End With
    LBALen = MSF2LBA(udtMSF)

    '/* look for the track in each drive,
    '/* as we can't rely on the drive letter
    '/* (ASPI...)
    For i = 0 To clsCDAccess.DeviceCount - 1
        If clsCDAccess.SelectDevice(i) Then
            clsCDAccess.GetTOC clsCurTOC

            With clsCurTOC
                For j = 1 To .TrackCount

                    If .track(j).StartLBA = LBA Then
                        If LBALen = (.track(j + 1).StartLBA - .track(j).StartLBA - 150) Then
                            Set clsCurTrack = .track(j)
                            lngLBAEnd = .track(j + 1).StartLBA - 1
                            intFoundTrack = 1
                        End If
                    End If

                    If intFoundTrack = 1 Then
                        Exit For
                    End If

                Next j
            End With

        End If

        If intFoundTrack = 1 Then
            Exit For
        End If

    Next i

    If intFoundTrack < 1 Then
        ISoundStream_StreamOpen = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    ISoundStream_StreamOpen = SND_ERR_SUCCESS
End Function

Private Function ISoundStream_StreamRead( _
    ByVal buffer_ptr As Long, _
    ByVal buffer_len As Long, _
    buffer_read As Long _
) As SND_RESULT

    buffer_read = 0

    Do While buffer_read < buffer_len
        If lngBufferLen - lngBufferPos < buffer_len - buffer_read Then

            If 0 < (lngBufferLen - lngBufferPos) Then
                CpyMem ByVal buffer_ptr + buffer_read, _
                       btBuffer(lngBufferPos), _
                       lngBufferLen - lngBufferPos

                buffer_read = buffer_read + (lngBufferLen - lngBufferPos)
            End If

            If Not FillBuffer Then
                ISoundStream_StreamRead = SND_ERR_END_OF_STREAM
                Exit Function
            End If

        Else

            CpyMem ByVal buffer_ptr + buffer_read, _
                   btBuffer(lngBufferPos), _
                   buffer_len - buffer_read

            lngBufferPos = lngBufferPos + (buffer_len - buffer_read)
            buffer_read = buffer_read + (buffer_len - buffer_read)

        End If
    Loop

    ISoundStream_StreamRead = SND_ERR_SUCCESS
End Function

Private Function ISoundStream_StreamSeek( _
    ByVal Value As Long, _
    ByVal seek_mode As SND_SEEK_MODE _
) As SND_RESULT

    Select Case seek_mode
        Case SND_SEEK_PERCENT
            If Value < 0 Or Value > 100 Then
                ISoundStream_StreamSeek = SND_ERR_OUT_OF_RANGE
                Exit Function
            End If
            lngLBA = Value * (ISoundInfo_Duration / 1000) / 100

        Case SND_SEEK_SECONDS
            If Value < 0 Or Value > (ISoundInfo_Duration / 1000) Then
                ISoundStream_StreamSeek = SND_ERR_OUT_OF_RANGE
                Exit Function
            End If
            lngLBA = (Value * (44100 * 4&)) / 2352&

    End Select

    ' force FillBuffer
    lngBufferLen = 0

    ISoundStream_StreamSeek = SND_ERR_SUCCESS
End Function
