VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "StreamOGG"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

' needs msvcrt.dll for C file access

Implements ISoundStream
Implements ISoundInfo


Private Const Extensions    As String = "ogg"
Private Const Description   As String = "Ogg Vorbis"


Private Type vorbis_info
    version                 As Long
    Channels                As Long
    rate                    As Long
    bitrate_upper           As Long
    bitrate_nominal         As Long
    bitrate_lower           As Long
    bitrate_window          As Long
    codec_setup             As Long
End Type

Private Enum byte_packaging
    bp_little_endian = 0
    bp_big_endian = 1
End Enum

Private Enum word_size
    word_size_8_bit = 1
    word_size_16_bit = 2
End Enum

Private Const BUFFER_SIZE   As Long = 4096

Private clsMSVCRT           As clsDLL
Private clsOgg              As clsDLL
Private clsVorbis           As clsDLL
Private clsVorbisFile       As clsDLL

Private udtInfo             As vorbis_info
Private vf(64& * 1024&)     As Byte

Private btBuffer()          As Byte

Private lngPosInBuffer      As Long
Private lngBufferData       As Long

Private lngDuration         As Long
Private lngCurPos           As Long

Private hFile               As Long
Private blnEOS              As Boolean

Private blnReady            As Boolean

Private clsTags             As StreamTags

Private Function ogg_init_libs( _
) As Boolean

    Dim bln As Boolean

    Set clsMSVCRT = New clsDLL
    Set clsOgg = New clsDLL
    Set clsVorbis = New clsDLL
    Set clsVorbisFile = New clsDLL

    bln = True

    bln = bln And clsMSVCRT.LoadDLL("msvcrt.dll")
    bln = bln And clsOgg.LoadDLL("ogg.dll")
    bln = bln And clsVorbis.LoadDLL("vorbis.dll")
    bln = bln And clsVorbisFile.LoadDLL("vorbisfile.dll")

    If Not bln Then Exit Function

    clsMSVCRT.IsCDECL = True
    clsOgg.IsCDECL = True
    clsVorbis.IsCDECL = True
    clsVorbisFile.IsCDECL = True

    ogg_init_libs = True
End Function

Private Sub ogg_free_libs()
    clsVorbisFile.UnloadDLL
    clsVorbis.UnloadDLL
    clsOgg.UnloadDLL
    clsMSVCRT.UnloadDLL
End Sub

''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''

Private Sub GetTags( _
    vfile() As Byte, _
    tags As StreamTags _
)

    Dim comment_ptr As Long
    Dim strTag      As String

    clsTags.ClearTags

    comment_ptr = ov_comment(VarPtr(vfile(0)), 0)
    If comment_ptr = 0 Then Exit Sub

    strTag = vorbis_comment_query(comment_ptr, "artist", 0)
    If strTag <> "" Then tags.AddTag "artist", strTag

    strTag = vorbis_comment_query(comment_ptr, "album", 0)
    If strTag <> "" Then tags.AddTag "album", strTag

    strTag = vorbis_comment_query(comment_ptr, "title", 0)
    If strTag <> "" Then tags.AddTag "title", strTag

    strTag = vorbis_comment_query(comment_ptr, "genre", 0)
    If strTag <> "" Then tags.AddTag "genre", strTag

    strTag = vorbis_comment_query(comment_ptr, "tracknumber", 0)
    If strTag <> "" Then tags.AddTag "tracknumber", strTag

    strTag = vorbis_comment_query(comment_ptr, "description", 0)
    If strTag <> "" Then tags.AddTag "comment", strTag
End Sub

Private Function FillBuffer( _
) As Boolean

    Dim lngCurrentSection   As Long
    Dim lngRet              As Long

    lngRet = ov_read(VarPtr(vf(0)), _
                     VarPtr(btBuffer(0)), _
                     BUFFER_SIZE, _
                     bp_little_endian, _
                     word_size_16_bit, _
                     True, _
                     lngCurrentSection)

    If lngRet = 0 Then
        blnEOS = True
    ElseIf lngRet > 0 Then
        lngCurPos = lngCurPos + lngRet * 1000 / udtInfo.rate / 2 / udtInfo.Channels
        lngBufferData = lngRet
        lngPosInBuffer = 0
        FillBuffer = True
    End If
End Function

''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''

Private Sub Class_Initialize()
    blnReady = ogg_init_libs

    Set clsTags = New StreamTags

    ReDim btBuffer(BUFFER_SIZE - 1) As Byte
End Sub

Private Sub Class_Terminate()
    ISoundStream_StreamClose
    ogg_free_libs
End Sub

''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''

Private Property Get ISoundInfo_BitsPerSample( _
) As Integer

    ISoundInfo_BitsPerSample = 16
End Property

Private Property Get ISoundInfo_BitsPerSecond( _
) As Long

    ISoundInfo_BitsPerSecond = udtInfo.bitrate_nominal
End Property

Private Property Get ISoundInfo_channels( _
) As Integer

    ISoundInfo_channels = udtInfo.Channels
End Property

Private Property Get ISoundInfo_Duration( _
) As Long

    ISoundInfo_Duration = lngDuration
End Property

Private Property Get ISoundInfo_Position( _
) As Long

    ISoundInfo_Position = lngCurPos
End Property

Private Property Get ISoundInfo_SamplesPerSecond( _
) As Long

    ISoundInfo_SamplesPerSecond = udtInfo.rate
End Property

Private Property Get ISoundInfo_Tags() As StreamTags
    Set ISoundInfo_Tags = clsTags
End Property

''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''

Private Property Get ISoundStream_Description( _
) As String

    ISoundStream_Description = Description
End Property

Private Property Get ISoundStream_EndOfStream( _
) As Boolean

    ISoundStream_EndOfStream = blnEOS
End Property

Private Function ISoundStream_Extensions( _
) As String()

    ISoundStream_Extensions = Split(Extensions, ";")
End Function

Private Function ISoundStream_StreamClose( _
) As SND_RESULT

    If hFile <> 0 Then
        ov_clear VarPtr(vf(0))
        clsMSVCRT.CallFunc "fclose", hFile
        hFile = 0
        blnEOS = False
        clsTags.ClearTags
        lngBufferData = 0
        lngPosInBuffer = 0

        ISoundStream_StreamClose = SND_ERR_SUCCESS
    Else
        ISoundStream_StreamClose = SND_ERR_INVALID_SOURCE
    End If
End Function

Private Property Get ISoundStream_StreamInfo( _
    ) As ISoundInfo

    Set ISoundStream_StreamInfo = Me
End Property

Private Function ISoundStream_QuickStreamInfo( _
    ByVal Source As String, _
    info As ISoundInfo _
) As SND_RESULT

    Dim hFileQ              As Long
    Dim btFile()            As Byte
    Dim btMode(2)           As Byte
    Dim vfQ(64& * 1024&)    As Byte
    Dim lngRet              As Long
    Dim clsInfo             As QuickStreamInfo
    Dim clsTagsQ            As StreamTags
    Dim udtInfoQ            As vorbis_info

    Set clsTagsQ = New StreamTags
    Set clsInfo = New QuickStreamInfo
    Set info = clsInfo

    If Not blnReady Then
        ISoundStream_QuickStreamInfo = SND_ERR_INTERNAL
        Exit Function
    End If

    If Not IsValidFile(Source) Then
        ISoundStream_QuickStreamInfo = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    btFile = StrConv(Source & Chr$(0), vbFromUnicode)

    btMode(0) = Asc("r")
    btMode(1) = Asc("b")

    hFileQ = clsMSVCRT.CallFunc("fopen", _
                                VarPtr(btFile(0)), _
                                VarPtr(btMode(0)))

    If hFileQ = 0 Then
        ISoundStream_QuickStreamInfo = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    If ov_open(hFileQ, VarPtr(vfQ(0)), 0, 0) < 0 Then
        ISoundStream_QuickStreamInfo = SND_ERR_INVALID_SOURCE
        clsMSVCRT.CallFunc "fclose", hFile
        Exit Function
    End If

    udtInfoQ = ov_info(VarPtr(vfQ(0)), -1)

    If udtInfoQ.rate < 1 Then
        ov_clear VarPtr(vfQ(0))
        clsMSVCRT.CallFunc "fclose", hFileQ
        ISoundStream_QuickStreamInfo = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    If udtInfoQ.Channels > 2 Then
        ov_clear VarPtr(vfQ(0))
        clsMSVCRT.CallFunc "fclose", hFileQ
        ISoundStream_QuickStreamInfo = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    ' ov_pcm_total returns a 64 bit value.
    ' We can only get a 32 bit value, so there shouldn't
    ' be more than 2 ^ 31 samples in the file,
    ' else we'd get the wrong duration.
    lngRet = ov_pcm_total(VarPtr(vfQ(0)), -1)

    GetTags vfQ, clsTagsQ

    With udtInfoQ
        clsInfo.SetValues .rate, _
                          .Channels, _
                          16, _
                          .bitrate_nominal, _
                          lngRet / udtInfoQ.rate * 1000, _
                          0, _
                          clsTagsQ
    End With

    ov_clear VarPtr(vfQ(0))
    clsMSVCRT.CallFunc "fclose", hFileQ

    ISoundStream_QuickStreamInfo = SND_ERR_SUCCESS
End Function

Private Function ISoundStream_StreamOpen( _
    ByVal Source As String _
) As SND_RESULT

    Dim btFile()    As Byte
    Dim btMode(2)   As Byte
    Dim lngRet      As Long

    ' blnReady gibt an, ob alle DLLs geladen
    ' werden konnten
    If Not blnReady Then
        ISoundStream_StreamOpen = SND_ERR_INTERNAL
        Exit Function
    End If

    ISoundStream_StreamClose

    If Not IsValidFile(Source) Then
        ISoundStream_StreamOpen = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    ' The file gets opened through the MS VC Runtime,
    ' VorbisFile doesn't acceppt other handles.
    ' We could use callbacks, but their calling
    ' convention is cdecl, so we'd need an asm bridge
    ' in there.
    btFile = StrConv(Source & Chr$(0), vbFromUnicode)

                            ' file...
    btMode(0) = Asc("r")    ' read
    btMode(1) = Asc("b")    ' binary

    hFile = clsMSVCRT.CallFunc("fopen", _
                               VarPtr(btFile(0)), _
                               VarPtr(btMode(0)))

    If hFile = 0 Then
        ISoundStream_StreamOpen = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    If ov_open(hFile, VarPtr(vf(0)), 0, 0) < 0 Then
        ISoundStream_StreamOpen = SND_ERR_INVALID_SOURCE
        clsMSVCRT.CallFunc "fclose", hFile
        Exit Function
    End If

    udtInfo = ov_info(VarPtr(vf(0)), -1)

    If udtInfo.rate < 1 Then
        ov_clear VarPtr(vf(0))
        clsMSVCRT.CallFunc "fclose", hFile
        ISoundStream_StreamOpen = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    If udtInfo.Channels > 2 Then
        ov_clear VarPtr(vf(0))
        clsMSVCRT.CallFunc "fclose", hFile
        ISoundStream_StreamOpen = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    ' ov_pcm_total returns a 64 bit value.
    ' We can only get a 32 bit value, so there shouldn't
    ' be more than 2 ^ 31 samples in the file,
    ' else we'd get the wrong duration.
    lngRet = ov_pcm_total(VarPtr(vf(0)), -1)

    lngDuration = lngRet / udtInfo.rate * 1000
    lngCurPos = 0

    GetTags vf, clsTags

    ISoundStream_StreamOpen = SND_ERR_SUCCESS
End Function

Private Function ISoundStream_StreamRead( _
    ByVal buffer_ptr As Long, _
    ByVal buffer_len As Long, _
    buffer_read As Long _
) As SND_RESULT

    If Not blnReady Then
        ISoundStream_StreamRead = SND_ERR_INTERNAL
        Exit Function
    End If

    ISoundStream_StreamRead = SND_ERR_SUCCESS

    buffer_read = 0

    Do While buffer_read < buffer_len

        If lngBufferData = 0 Then
            If Not FillBuffer Then
                ISoundStream_StreamRead = SND_ERR_END_OF_STREAM
                Exit Function
            End If

        ElseIf (lngBufferData - lngPosInBuffer) < (buffer_len - buffer_read) Then
            If 0 < (lngBufferData - lngPosInBuffer) Then

                If 0 = IsBadReadPtr(ByVal VarPtr(btBuffer(0)) + lngPosInBuffer, _
                                    lngBufferData - lngPosInBuffer) Then

                    If 0 = IsBadWritePtr(ByVal buffer_ptr + buffer_read, _
                                         lngBufferData - lngPosInBuffer) Then

                        CpyMem ByVal buffer_ptr + buffer_read, _
                               ByVal VarPtr(btBuffer(0)) + lngPosInBuffer, _
                               lngBufferData - lngPosInBuffer

                    End If
                End If

                buffer_read = buffer_read + (lngBufferData - lngPosInBuffer)
            End If

            If Not FillBuffer Then
                ISoundStream_StreamRead = SND_ERR_END_OF_STREAM
                Exit Function
            End If

        Else
            If 0 = IsBadReadPtr(ByVal VarPtr(btBuffer(0)) + lngPosInBuffer, _
                                buffer_len - buffer_read) Then

                If 0 = IsBadWritePtr(ByVal buffer_ptr + buffer_read, _
                                     buffer_len - buffer_read) Then

                    CpyMem ByVal buffer_ptr + buffer_read, _
                           ByVal VarPtr(btBuffer(0)) + lngPosInBuffer, _
                           buffer_len - buffer_read

                End If
            End If

            lngPosInBuffer = lngPosInBuffer + (buffer_len - buffer_read)
            buffer_read = buffer_read + (buffer_len - buffer_read)

        End If

    Loop
End Function

Private Function ISoundStream_StreamSeek( _
    ByVal Value As Long, _
    ByVal seek_mode As SND_SEEK_MODE _
) As SND_RESULT

    Dim dblTime     As Double

    If Not blnReady Then
        ISoundStream_StreamSeek = SND_ERR_INTERNAL
        Exit Function
    End If

    Select Case seek_mode

        Case SND_SEEK_PERCENT
            If Value < 0 Or Value > 99 Then
                ISoundStream_StreamSeek = SND_ERR_OUT_OF_RANGE
                Exit Function
            End If

            dblTime = (ISoundInfo_Duration / 1000) / 100 * Value

        Case SND_SEEK_SECONDS
            If Value < 0 Or Value > (ISoundInfo_Duration / 1000) Then
                ISoundStream_StreamSeek = SND_ERR_OUT_OF_RANGE
                Exit Function
            End If

            dblTime = Value

    End Select

    lngCurPos = dblTime * 1000
    ov_time_seek VarPtr(vf(0)), dblTime

    lngBufferData = 0
    lngPosInBuffer = 0

    blnEOS = False

    ISoundStream_StreamSeek = SND_ERR_SUCCESS
End Function

''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''

Private Function vorbis_comment_query( _
    ByVal vorbis_comment As Long, _
    ByVal TagName As String, _
    ByVal count As Long _
) As String

    Dim btTag() As Byte
    Dim tag_ptr As Long

    If vorbis_comment <> 0 Then
        btTag = StrConv(TagName & Chr$(0), vbFromUnicode)
        tag_ptr = clsVorbis.CallFunc("vorbis_comment_query", _
                                     vorbis_comment, _
                                     VarPtr(btTag(0)), _
                                     0)
        If tag_ptr <> 0 Then
            vorbis_comment_query = Ptr2Str(tag_ptr)
        End If
    End If
End Function

Private Function ov_comment( _
    ByVal ovf_struct As Long, _
    ByVal link As Long _
) As Long

    ov_comment = clsVorbisFile.CallFunc("ov_comment", ovf_struct, link)
End Function

Private Function ov_time_seek( _
    ByVal ovf_struct As Long, _
    ByVal seconds As Double _
) As Long

    Dim lngTimeLo   As Long
    Dim lngTimeHi   As Long

    CpyMem lngTimeLo, seconds, 4
    CpyMem lngTimeHi, ByVal VarPtr(seconds) + 4, 4

    ov_time_seek = clsVorbisFile.CallFunc("ov_time_seek", _
                                          ovf_struct, _
                                          lngTimeLo, _
                                          lngTimeHi)
End Function

Private Function ov_read( _
    ByVal ovf_struct As Long, _
    ByVal output_ptr As Long, _
    ByVal Length As Long, _
    ByVal bigendianp As byte_packaging, _
    ByVal word As word_size, _
    ByVal signed As Boolean, _
    ByRef bitstream As Long _
) As Long

    ov_read = clsVorbisFile.CallFunc("ov_read", _
                                     ovf_struct, _
                                     output_ptr, _
                                     Length, _
                                     bigendianp, _
                                     word, _
                                     Abs(signed), _
                                     VarPtr(bitstream))
End Function

Private Function ov_pcm_total( _
    ByVal ovf_struct As Long, _
    ByVal link As Long _
) As Long

    ov_pcm_total = clsVorbisFile.CallFunc("ov_pcm_total", _
                                          ovf_struct, _
                                          link)
End Function

Private Function ov_info( _
    ByVal ovf_struct As Long, _
    ByVal link As Long _
) As vorbis_info

    Dim ptr     As Long
    Dim udt     As vorbis_info

    ptr = clsVorbisFile.CallFunc("ov_info", ovf_struct, link)
    If ptr = 0 Then Exit Function

    CpyMem udt, ByVal ptr, Len(udt)

    ov_info = udt
End Function

Private Function ov_open( _
    ByVal hFile As Long, _
    ByVal ovf_struct As Long, _
    ByVal initial_ptr As Long, _
    ByVal ibytes As Long _
) As Long

    ov_open = clsVorbisFile.CallFunc("ov_open", _
                                     hFile, _
                                     ovf_struct, _
                                     initial_ptr, _
                                     ibytes)
End Function

Private Sub ov_clear( _
    ByVal ovf_struct As Long _
)

    clsVorbisFile.CallFunc "ov_clear", ovf_struct
End Sub
