Attribute VB_Name = "modDSP"
Option Explicit

Private Enum VOL_UNIT
    VOL_DECIBEL = 0
    VOL_PERCENT
    VOL_FACTOR
End Enum

Private Const Pi                    As Single = 3.14159265358979

Private intEcho16()                 As Integer
Private lngEchoPos16                As Long
Private sngEchoLength16             As Single

Public Const EQU_BANDS              As Long = 8
Public clsEQUBands(EQU_BANDS - 1)   As IIRFilter

Private sngShX1                     As Single
Private sngShX2                     As Single
Private sngShX3                     As Single
Private sngShX4                     As Single

Private sngShY1                     As Single
Private sngShY2                     As Single
Private sngShY3                     As Single
Private sngShY4                     As Single

Private sngShMinWP                  As Single
Private sngShMaxWP                  As Single
Private sngShWp                     As Single
Private sngShRate                   As Single
Private sngShSweepFac               As Single

Private clsSmNorm(1)                As clsSmooth

Private lngSamplerate               As Long
Private intChannels                 As Integer

Public Sub InitDSP()
    Dim i   As Long

    For i = 0 To EQU_BANDS - 1
        Set clsEQUBands(i) = New IIRFilter
    Next
End Sub

Public Sub DeinitDSP()
    '
End Sub

Private Sub UpdateEQU()
    Dim i           As Long
    Dim lngFreq     As Long
    Dim sngBW       As Single
    Dim sngBFreq    As Single

    For i = 0 To EQU_BANDS - 1
        lngFreq = lngSamplerate / 3

        sngBW = Log(lngFreq / 80#) / 4

        With clsEQUBands(i)
            sngBFreq = 80# * (lngFreq / 80#) ^ (i / (EQU_BANDS - 1))
            .CreateBiquadIIR IIR_PEAK_EQ, .GainDB, sngBFreq, lngSamplerate, sngBW
            .Enabled = settingEQUEnabled
        End With
    Next

    frmEqualizer.Update
End Sub

Public Sub UpdateEcho()
    Dim lngEchoPoints   As Long

    If lngSamplerate = 0 Then Exit Sub

    lngEchoPoints = lngSamplerate / 1000 * settingEchoLen
    ReDim intEcho16(lngEchoPoints - 1) As Integer

    sngEchoLength16 = settingEchoAmp
    lngEchoPos16 = 0
End Sub

Public Sub UpdateNormalization()
    Dim i   As Long

    If clsSmNorm(0) Is Nothing Then
        Set clsSmNorm(0) = New clsSmooth
        Set clsSmNorm(1) = New clsSmooth
    End If

    clsSmNorm(0).SmoothNew 100
    clsSmNorm(1).SmoothNew 100
End Sub

' http://www.codeproject.com/cs/media/cswavplayfx.asp
Public Sub UpdatePhaseShift()
    Dim range   As Double

    If lngSamplerate = 0 Then Exit Sub

    sngShMinWP = Pi * settingShFreq / lngSamplerate
    range = 2 ^ settingShSweepRange
    sngShMaxWP = Pi * settingShFreq * range / lngSamplerate
    sngShRate = range ^ (2 * settingShSweepRate / lngSamplerate)

    sngShSweepFac = sngShRate
    sngShWp = sngShMinWP
End Sub

Public Sub FormatChanged( _
    ByVal samplerate As Long, _
    ByVal Channels As Integer, _
    ByVal BitsPerSample As Integer _
)

    lngSamplerate = samplerate
    intChannels = Channels

    UpdateEQU
    UpdateEcho
    UpdatePhaseShift
    UpdateNormalization
End Sub

Public Sub ProcessData( _
    ByVal data_ptr As Long, _
    ByVal data_len As Long _
)

    Dim i           As Long
    Dim intData()   As Integer

    If data_len = 0 Then Exit Sub

    ReDim intData(data_len / 2 - 1) As Integer
    CpyMem intData(0), ByVal data_ptr, data_len

    ' Normalization
    If (settingFXFlags And FX_NORMAL) = FX_NORMAL Then
        DSP_Process_Normalization intData
    End If

    ' Equalizer
    If settingEQUEnabled Then
        DSP_Process_ChangeVolume intData, settingPreamp, VOL_DECIBEL

        For i = 0 To EQU_BANDS - 1
            clsEQUBands(i).ProcessSamples_16bit intData
        Next
    End If

    ' Phase Shift
    If (settingFXFlags And FX_PHSHFT) = FX_PHSHFT Then
        DSP_Process_PhaseShift intData
    End If

    ' Channel Swap (only stereo)
    If (settingFXFlags And FX_CHSWAP) = FX_CHSWAP Then
        DSP_Process_ChSwp intData
    End If

    ' Reverb
    If (settingFXFlags And FX_REVERB) = FX_REVERB Then
        DSP_Process_Echo intData
    End If

    CpyMem ByVal data_ptr, intData(0), data_len
End Sub

''''''''''''''''''''''''''''''''''''''''''''''''''''''
' translated from the XMMS volume normalization plugin
' http://volnorm.sourceforge.net/

Private Sub DSP_Process_Normalization( _
    intSamples() As Integer _
)

    Const max_mult      As Single = 1

    Dim level           As Single
    Dim channel         As Integer
    Dim ch_level        As Single
    Dim gain            As Single

    calc_power_level intSamples

    level = -1

    For channel = 0 To intChannels - 1
        ch_level = clsSmNorm(channel).SmoothGetMax
        If ch_level > level Then level = ch_level
    Next

    If level = 0 Then level = 1
    gain = settingVolNorm / level

    If gain > max_mult Then
        gain = max_mult
    End If

    adjust_gain intSamples, gain
End Sub

Private Sub adjust_gain(intSamples() As Integer, ByVal gain As Single)
    Const NO_GAIN   As Single = 0.01

    Dim i           As Long
    Dim lngTemp     As Long

    If (gain >= 1# - NO_GAIN And gain <= 1# + NO_GAIN) Then
        Exit Sub
    End If

    For i = 0 To UBound(intSamples)
        lngTemp = intSamples(i) * gain
        If lngTemp > 32767 Then
            intSamples(i) = 32767
        ElseIf lngTemp < -32768 Then
            intSamples(i) = -32768
        Else
            intSamples(i) = CInt(lngTemp)
        End If
    Next
End Sub

Private Sub calc_power_level(intSamples() As Integer)
    Const NORMAL    As Single = 1 / 32767
    Const NORMAL_SQ As Single = NORMAL * NORMAL

    Dim channel     As Integer
    Dim i           As Long
    Dim sum(1)      As Single
    Dim Length      As Long
    Dim sample      As Single
    Dim temp        As Single
    Dim ch_len      As Single
    Dim level       As Single

    For i = 0 To UBound(intSamples)
        sample = intSamples(i)
        temp = sample * sample

        sum(channel) = sum(channel) + temp

        channel = (channel + 1) Mod intChannels
    Next

    ch_len = 1 / (UBound(intSamples) + 1)

    For channel = 0 To intChannels - 1
        level = sum(channel) * ch_len * NORMAL_SQ
        clsSmNorm(channel).SmoothAdd Sqr(level)
    Next
End Sub

'
''''''''''''''''''''''''''''''''''''''''''''''''''

Private Sub DSP_Process_ChangeVolume( _
    intSamples() As Integer, _
    ByVal Value As Single, _
    ByVal unit As VOL_UNIT _
)

    Dim sngFactor   As Single
    Dim sngResult   As Single
    Dim i           As Long

    Select Case unit
        Case VOL_DECIBEL
            sngFactor = 10 ^ (Value / 20)
        Case VOL_PERCENT
            sngFactor = Value / 100
        Case VOL_FACTOR
            sngFactor = Value
    End Select

    For i = 0& To UBound(intSamples)
        sngResult = intSamples(i) * sngFactor

        If sngResult > 32767# Then
            intSamples(i) = 32767
        ElseIf sngResult < -32768# Then
            intSamples(i) = -32768
        Else
            intSamples(i) = CInt(sngResult)
        End If
    Next
End Sub

Private Sub DSP_Process_ChSwp( _
    intSamples() As Integer _
)

    Dim i       As Long
    Dim intCh   As Integer

    If intChannels <> 2 Then Exit Sub

    For i = 0 To UBound(intSamples) Step 2
        intCh = intSamples(i)
        intSamples(i) = intSamples(i + 1)
        intSamples(i + 1) = intCh
    Next
End Sub

Public Sub DSP_Process_Echo( _
    intSamples() As Integer _
)

    Dim i   As Long

    For i = 0 To UBound(intSamples)
        intSamples(i) = norm16(CLng(intSamples(i)) + intEcho16(lngEchoPos16))
        intEcho16(lngEchoPos16) = intSamples(i) * sngEchoLength16

        lngEchoPos16 = lngEchoPos16 + 1
        If lngEchoPos16 > UBound(intEcho16) Then
            lngEchoPos16 = 0
        End If
    Next
End Sub

' http://www.codeproject.com/cs/media/cswavplayfx.asp
Public Sub DSP_Process_PhaseShift( _
    intSamples() As Integer _
)

    Dim i   As Long
    Dim k   As Single
    Dim X1  As Single
    Dim X   As Single
    Dim Y   As Single

    For i = 0& To UBound(intSamples)
        X = intSamples(i) / 32767

        k = (1# - sngShWp) / (1# + sngShWp)

        X1 = X + settingShFeedback * sngShY4
        sngShY1 = k * (sngShY1 + X1) - sngShX1
        sngShX1 = X1
        sngShY2 = k * (sngShY2 + sngShY1) - sngShX2
        sngShX2 = sngShY1
        sngShY3 = k * (sngShY3 + sngShY2) - sngShX3
        sngShX3 = sngShY2
        sngShY4 = k * (sngShY4 + sngShY3) - sngShX4
        sngShX4 = sngShY3

        Y = sngShY4 * settingShWet + X * settingShDry

        sngShWp = sngShWp * sngShSweepFac
        If (sngShWp > sngShMaxWP) Then
            sngShSweepFac = 1# / sngShRate
        ElseIf (sngShWp < sngShMinWP) Then
            sngShSweepFac = sngShRate
        End If

        Y = Y * 32767

        If Y > 32767 Then
            intSamples(i) = 32767
        ElseIf Y < -32768 Then
            intSamples(i) = -32768
        Else
            intSamples(i) = CInt(Y)
        End If
    Next
End Sub

Private Function norm16( _
    ByVal dbl As Single _
) As Integer

    If dbl > 32767 Then
        norm16 = 32767
    ElseIf dbl < -32768 Then
        norm16 = -32768
    Else
        norm16 = CInt(dbl)
    End If
End Function
