Attribute VB_Name = "modMain"
Option Explicit

Private Declare Function InitCommonControls Lib "Comctl32.dll" ( _
) As Long

Public Enum REPEAT_MODE
    REPEAT_ALL = 0
    REPEAT_SINGLE
    REPEAT_NONE
End Enum

Public Enum EFFECT_FLAGS
    FX_REVERB = 2 ^ 0
    FX_CHSWAP = 2 ^ 1
    FX_PHSHFT = 2 ^ 2
    FX_NORMAL = 2 ^ 3
End Enum

Public Const APPNAME        As String = "Monoton"

Public settingDSDeviceGUID  As String
Public settingDSSamplerate  As Long
Public settingDSStereo      As Boolean
Public settingRepeatMode    As REPEAT_MODE
Public settingShuffle       As Boolean
Public settingFXFlags       As EFFECT_FLAGS
Public settingVolDB         As Long
Public settingPan           As Long
Public settingCDText        As Boolean
Public settingEQUEnabled    As Boolean
Public settingPreamp        As Long
Public settingEchoLen       As Long
Public settingEchoAmp       As Single
Public settingShDry         As Single
Public settingShWet         As Single
Public settingShFeedback    As Single
Public settingShSweepRate   As Single
Public settingShSweepRange  As Single
Public settingShFreq        As Single
Public settingVolNorm       As Single
Public settingFreeDBUser    As String
Public settingFreeDBUserh   As String
Public settingFreeDBAgent   As String
Public settingFreeDBAgentv  As String
Public settingEncCfgWAV     As String
Public settingEncCfgMP3     As String
Public settingEncCfgAPE     As String
Public settingEncCfgWMA     As String
Public settingEncCfgOGG     As String

Sub Main()
    Dim i           As Long

    GetSettings

    ' CDDA support
    modCDRom.InitCDRomSupport

    For i = 0 To modCDRom.CDInterfaceCount - 1
        If modCDRom.CDSelectInterface(i) Then Exit For
    Next

    If clsCDAccess.InterfaceName = "" Then
        MsgBox "ASPI/SPTI/IOCTL initialization failed!" & vbCrLf & _
               "CD-ROM support disabled!", vbExclamation, APPNAME
    End If

    ' DirectSound 8
    If Not modDirectX.InitializeDirectX Then
        MsgBox "Could not init DirectX!", vbExclamation, APPNAME
        Exit Sub
    End If

    For i = 1 To modDirectX.DS_DeviceCount
        If StrComp(modDirectX.DS_DeviceGUID(i), settingDSDeviceGUID, vbTextCompare) = 0 Then
            Exit For
        End If
    Next

    If i > modDirectX.DS_DeviceCount Then
        settingDSDeviceGUID = modDirectX.DS_DeviceGUID(1)
    End If

    If Not modDirectX.InitDirectSound(settingDSSamplerate, _
                                      IIf(settingDSStereo, 2, 1), _
                                      16, _
                                      settingDSDeviceGUID) Then

        MsgBox "Couldn't init DirectSound!", vbExclamation, APPNAME
        Exit Sub
    End If

    If Not IsProgIDRegistered("SSubTimer6.CTimer") Then
        RegisterCOMServer AddSlash(App.path) & "SSubTmr6.dll"
    End If

    If Not IsProgIDRegistered("vbAcceleratorSGrid6.vbalGrid") Then
        RegisterCOMServer AddSlash(App.path) & "vbalSGrid6.ocx"
    End If

    LoadLibrary "shell32.dll"   ' should fix visual styles shutdown crashes
    InitCommonControls

    frmMain.Show
End Sub

Public Sub SaveSettings()
    SaveSetting APPNAME, "settings", "DeviceGUID", settingDSDeviceGUID
    SaveSetting APPNAME, "settings", "RepeatMode", settingRepeatMode
    SaveSetting APPNAME, "settings", "Shuffle", Abs(settingShuffle)
    SaveSetting APPNAME, "settings", "FXFlags", settingFXFlags
    SaveSetting APPNAME, "settings", "Volume", settingVolDB
    SaveSetting APPNAME, "settings", "CDText", settingCDText
    SaveSetting APPNAME, "settings", "Panning", settingPan
    SaveSetting APPNAME, "settings", "EQUEnabled", settingEQUEnabled
    SaveSetting APPNAME, "settings", "Preamp", settingPreamp
    SaveSetting APPNAME, "settings", "EchoLen", settingEchoLen
    SaveSetting APPNAME, "settings", "EchoAmp", settingEchoAmp
    SaveSetting APPNAME, "settings", "ShDry", settingShDry
    SaveSetting APPNAME, "settings", "ShWet", settingShWet
    SaveSetting APPNAME, "settings", "ShFb", settingShFeedback
    SaveSetting APPNAME, "settings", "ShSwpRate", settingShSweepRate
    SaveSetting APPNAME, "settings", "ShSwpRange", settingShSweepRange
    SaveSetting APPNAME, "settings", "ShFreq", settingShFreq
    SaveSetting APPNAME, "settings", "VolNorm", settingVolNorm
    SaveSetting APPNAME, "settings", "DSSamplerate", settingDSSamplerate
    SaveSetting APPNAME, "settings", "DSStereo", settingDSStereo
    SaveSetting APPNAME, "settings", "FreeDBUser", settingFreeDBUser
    SaveSetting APPNAME, "settings", "FreeDBUserHost", settingFreeDBUserh
    SaveSetting APPNAME, "settings", "FreeDBAgent", settingFreeDBAgent
    SaveSetting APPNAME, "settings", "FreeDBAgentVer", settingFreeDBAgentv
    SaveSetting APPNAME, "settings", "EncAPE", settingEncCfgAPE
    SaveSetting APPNAME, "settings", "EncMP3", settingEncCfgMP3
    SaveSetting APPNAME, "settings", "EncOGG", settingEncCfgOGG
    SaveSetting APPNAME, "settings", "EncWAV", settingEncCfgWAV
    SaveSetting APPNAME, "settings", "EncWMA", settingEncCfgWMA
End Sub

Private Sub GetSettings()
    settingDSDeviceGUID = CStr(GetSetting(APPNAME, "settings", "DeviceGUID", ""))
    settingRepeatMode = CInt(GetSetting(APPNAME, "settings", "RepeatMode", 0))
    settingShuffle = CBool(GetSetting(APPNAME, "settings", "Shuffle", 0))
    settingFXFlags = CInt(GetSetting(APPNAME, "settings", "FXFlags", 0))
    settingVolDB = CLng(GetSetting(APPNAME, "settings", "Volume", 0))
    settingCDText = CBool(GetSetting(APPNAME, "settings", "CDText", 0))
    settingPan = CLng(GetSetting(APPNAME, "settings", "Panning", 0))
    settingEQUEnabled = CBool(GetSetting(APPNAME, "settings", "EQUEnabled", 0))
    settingPreamp = CLng(GetSetting(APPNAME, "settings", "Preamp", 0))
    settingEchoLen = CLng(GetSetting(APPNAME, "settings", "EchoLen", 100))
    settingEchoAmp = CSng(GetSetting(APPNAME, "settings", "EchoAmp", 0.4))
    settingShDry = CSng(GetSetting(APPNAME, "settings", "ShDry", 1#))
    settingShWet = CSng(GetSetting(APPNAME, "settings", "ShWet", 1#))
    settingShFeedback = CSng(GetSetting(APPNAME, "settings", "ShFb", 0.5))
    settingShSweepRate = CSng(GetSetting(APPNAME, "settings", "ShSwpRate", 1#))
    settingShSweepRange = CSng(GetSetting(APPNAME, "settings", "ShSwpRange", 4#))
    settingShFreq = CSng(GetSetting(APPNAME, "settings", "ShFreq", 100))
    settingVolNorm = CSng(GetSetting(APPNAME, "settings", "VolNorm", 0.25))
    settingDSSamplerate = CLng(GetSetting(APPNAME, "settings", "DSSamplerate", 44100))
    settingDSStereo = CBool(GetSetting(APPNAME, "settings", "DSStereo", True))
    settingFreeDBUser = CStr(GetSetting(APPNAME, "settings", "FreeDBUser", "rm"))
    settingFreeDBUserh = CStr(GetSetting(APPNAME, "settings", "FreeDBUserHost", "rm_code.dl.am"))
    settingFreeDBAgent = CStr(GetSetting(APPNAME, "settings", "FreeDBAgent", "Monoton"))
    settingFreeDBAgentv = CStr(GetSetting(APPNAME, "settings", "FreeDBAgentVer", "2.0"))
    settingEncCfgAPE = CStr(GetSetting(APPNAME, "settings", "EncAPE", ""))
    settingEncCfgMP3 = CStr(GetSetting(APPNAME, "settings", "EncMP3", ""))
    settingEncCfgOGG = CStr(GetSetting(APPNAME, "settings", "EncOGG", ""))
    settingEncCfgWAV = CStr(GetSetting(APPNAME, "settings", "EncWAV", ""))
    settingEncCfgWMA = CStr(GetSetting(APPNAME, "settings", "EncWMA", ""))
End Sub
